/*	Justin Fiedler
	main.cpp
	This program accepts 3 command line arguments <option> <input image> <output image>.  It loads the input image and
	depending on which option (0-4) if selected the image is either 0)Desaturated 1)3x3 B-splined 2)3x3 Boxed
	3)5x5 B-splined or 4)5x5 Boxed. The output is then save to the file name given by the user.
*/

#include <iostream>
using namespace std;
#include "TargaImage.h"

TargaImage* img_mod;	//modified image
TargaImage* img_in;		//image from user input
char* img_out;			//name of file to save image to
int opt;				//option selected by user
bool s_flag = true;		//flag to control saving the file

//convolution filters
unsigned char bspline3[] = {1, 2, 1,
							2, 4, 2,
							1, 2, 1};

unsigned char box3[] = {1, 1, 1,
						1, 1, 1,
						1, 1, 1};

unsigned char bspline5[] = {1,  4,  6,  4, 1,
							4, 16, 24, 16, 4,
							6, 24, 36, 24, 6,
							4, 16, 24, 16, 4,
							1,  4,  6,  4, 1,};

unsigned char box5[] = {1, 1, 1, 1, 1,
						1, 1, 1, 1, 1,
						1, 1, 1, 1, 1,
						1, 1, 1, 1, 1,
						1, 1, 1, 1, 1};

/* returns pixel <x>,<y> from <image> or NULL if location is outside of the image
*/
unsigned char* getPixel(TargaImage* image, int x, int y){
	if(x < 0 || x >= image->width || y < 0 || y >= image->height) {
		return NULL;
	} else {
		return image->data + (y * img_in->width + x) * 4;
	}
}

/* desaturates <img_in> and puts it into <img_mod>
*/
void desaturate() {
	//loop through the image and replace image pixels desaturated ones;
	for(int x=0; x < img_in->width; x++){
		for(int y=0; y < img_in->height; y++) {
			unsigned char* pxl_in  = getPixel(img_in, x,y);
			unsigned char* pxl_mod = getPixel(img_mod,x,y);
				int L = (pxl_in[0] + pxl_in[1] + pxl_in[2])/3;
				pxl_mod[0]	= .7 * pxl_in[0] + .3 * L;  // desaturate red
				pxl_mod[1]	= .7 * pxl_in[1] + .3 * L;	// desaturate green
				pxl_mod[2]	= .7 * pxl_in[2] + .3 * L;	// desaturate blue
				pxl_mod[3]	= pxl_in[3];				// leave alpha alone

		}
	}
}

/* returns the norm factor of a given pixel from <image> and the given <kernel>
*/
int getNormFactor(TargaImage* image, const int x, const int y, unsigned char* kernel, int size){
	unsigned char* temp;
	int norm_factor = 0;
	for(int i = 0; i < size; i++){
		for(int j = 0; j < size; j++){
			if((temp = getPixel(image, x + (i - size/2), y + (j - size/2))) != NULL){
				norm_factor += kernel[i * size + j];
			}
		}
	}
	return norm_factor;
}

/* convolutes pixel <x>,<y> in <img_mod> given <img_in> and <kernel> of <size>
*/
void convolute_pxl(int x, int y, unsigned char* kernel, int size){
	//get and clear old pixel info
	unsigned char* pxl_mod = getPixel(img_mod, x, y);
	for(int c = 0; c < 4; c++){
		pxl_mod[c] = 0;
	}
	unsigned char* temp;

	//get norm factor
	int norm_factor = getNormFactor(img_in, x, y, kernel, size);
	
	for(int i = 0; i < size; i++){				//loop over size of filter
		for(int j = 0; j < size; j++){
			if((temp = getPixel(img_in, x + (i - size/2), y + (j - size/2))) != NULL){	//if the pixel exists add its weighted value to the new pixel
				for(int c = 0; c < 4; c++){		//modify each component
					pxl_mod[c] += kernel[i * size + j] * temp[c] / norm_factor;
				}
			}
		}
	}
}

/* convolutes <img_mod> given <img_in> and <kernel> of <size>
*/
void convolute(unsigned char* kernel, int size) {
	//convolute each pixel over image
	for(int x=0; x < img_in->width; x++){
		for(int y=0; y < img_in->height; y++) {
			convolute_pxl(x, y, kernel, size);
		}
	}
}

int main(int argc, char** args)
{	if(argc == 4){
		//get input values
		opt = atoi(args[1]);
		img_in = TargaImage::Load_Image(args[2]);
		img_mod = TargaImage::Load_Image(args[2]);
		img_out = args[3];
		if(img_in == NULL || img_mod == NULL) {
			cerr << "Failed to load input image." << endl;
			s_flag = false;
		}

		//choose an option
		switch(opt) {
			case 0: //Desaturate
				cout << "Desaturate Image...";
				desaturate();
				cout << "complete" << endl;
				break;
			case 1: // 3x3 B-Spline
				cout << "3x3 B-Spline Convolution...";
				convolute(bspline3, 3);
				cout << "complete" << endl;
				break;
			case 2: //3x3 Box 
				cout << "3x3 Box Convolution...";
				convolute(box3, 3);
				cout << "complete" << endl;
				break;
			case 3:	//5x5 BSpline
				cout << "5x5 B-Spline Convolution...";
				convolute(bspline5, 5);
				cout << "complete" << endl;
				break;
			case 4: //5x5 Box
				cout << "5x5 Box Convolution...";
				convolute(box5, 5);
				cout << "complete" << endl;
				break;
			default:
				cerr << "Please enter a valid option(0-4)" << endl;
				s_flag = false;
		}
		
		//save image
		if(s_flag){
			if(!img_mod->Save_Image(img_out)){
				cerr << "Image failed to save to " << img_out << endl;
			}
		}
		/* run all at once (sometimes i'm too lazy for command line)
		cout << "Desaturate Image..."; desaturate(); cout << "complete" << endl; 
		if(!img_mod->Save_Image("desaturate.tga")) cout << "save failed" << endl;
		cout << "3x3 B-Spline Convolution..."; convolute(bspline3, 3); cout << "complete" << endl;
		if(!img_mod->Save_Image("bspline3.tga")) cout << "save failed" << endl;
		cout << "3x3 Box Convolution..."; convolute(box3, 3); cout << "complete" << endl;
		if(!img_mod->Save_Image("box3.tga")) cout << "save failed" << endl;
		cout << "5x5 B-Spline Convolution..."; convolute(bspline5, 5); cout << "complete" << endl;
		if(!img_mod->Save_Image("bspline5.tga")) cout << "save failed" << endl;
		cout << "5x5 Box Convolution..."; convolute(box5, 5); cout << "complete" << endl;
		if(!img_mod->Save_Image("box5.tga")) cout << "save failed" << endl;
		*/
		
	} else { //recite error in case you befuddle the command line
		cerr << "usage p2: <option> <input image> <output image>";
	}

	
}